use crate::prelude::*;

use serde::{Deserialize, Serialize};
use std::{cell::RefCell, rc::Rc};

#[derive(Default)]
pub struct LocalDiscovery {
    targets: RefCell<Vec<fleck_core::peer::Peer>>,
}

#[derive(Serialize, Deserialize, Debug)]
struct DiscoveryMsg {
    pkey: fleck_core::crypto::PublicKey,
}

impl fleck_core::msg::MessageParams for DiscoveryMsg {
    const NAME: &'static str = "DiscoveryMsg";
    const ENCRYPTED: bool = false;
}

impl Service for Rc<LocalDiscovery> {
    fn setup(&self, api: &crate::Fleck) {
        api.with_service(|m: &fleck_core::MessageService| {
            m.add_message_type::<DiscoveryMsg>(api);
        });
        api.channel::<fleck_core::MajorTickChannel>()
            .sub_ref(self, LocalDiscovery::process_major_tick);
        api.channel::<fleck_core::msg::MessageChannel<DiscoveryMsg>>()
            .sub_eat(self, LocalDiscovery::discovery);
    }
}

impl LocalDiscovery {
    pub fn add_multicast_peer(&self, peer: fleck_core::peer::Peer) {
        self.targets.borrow_mut().push(peer);
    }

    fn discovery(&self, api: &crate::Fleck, msg: (fleck_core::msg::Metadata, DiscoveryMsg)) {
        let discovery = msg.1;
        log::trace!("received discovery message");

        // this will automatically ignore self-messages because we already know the pubkey
        api.with_service(|ns: &fleck_core::NodeService| {
            if ns.node_by_pubkey(&discovery.pkey).is_none() {
                // let node = fleck_core::Node::build_with_peer(.data.addr.unwrap());
                // node.set_pubkey(discovery.pkey);
                // nodes.inform_of(node);
            }
        });
    }

    /*fn packet(&self, api: &crate::Fleck, pkt: fleck_core::msg::Message) -> Option<fleck_core::msg::Message> {
        /*if let Some(discovery) = pkt.data.as_msg::<DiscoveryMsg>() {
            log::trace!("received discovery message");

            // this will automatically ignore self-messages because we already know the pubkey
            api.with_service(|nodes: &fleck_core::Nodes| {
                if nodes.node_by_pubkey(&discovery.pkey).is_none() {
                    let node = fleck_core::Node::build_with_addr(pkt.data.addr.unwrap());
                    node.set_pubkey(discovery.pkey);
                    nodes.inform_of(node);
                }
            });
        }*/
    }*/

    fn process_major_tick(&self, api: &crate::Fleck, _: &mut ()) {
        for peer in self.targets.borrow().iter() {
            let msg = fleck_core::msg::Message::build(DiscoveryMsg {
                pkey: api.with_service(|ns: &fleck_core::NodeService| {
                    ns.self_node().pubkey().unwrap().to_owned()
                }),
            })
            .with_peer(peer.clone());

            api.queue::<fleck_core::SendPacketChannel>(msg);
        }
        /*api.queue::<fleck_core::SendPacketChannel>(FleckPacket {
            addr: Some((*crate::io::MULTICAST_ADDRESS).into()),
            data: None,
            io_channel: Some(api.raw_io().local().clone()),
            node: None,
            msg: Some(crate::msg::Message::build(DiscoveryMsg {
                pkey: api.with_service(|nodes: &fleck_core::Nodes| nodes.self_node().pubkey().unwrap().to_owned())
            })),
        });*/
    }
}
